@@ -1,4 +1,4 @@
-# Copyright 2004-2015 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2017 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -30,10 +30,11 @@ import renpy.text.font as font
 import renpy.text.extras as extras
 
 try:
-    from _renpybidi import log2vis, WRTL, RTL, ON
+    from _renpybidi import log2vis, WRTL, RTL, ON  # @UnresolvedImport
 except:
     pass
 
+
 class Blit(object):
     """
     Represents a blit command, which can be used to render a texture to a
@@ -97,7 +98,6 @@ def outline_blits(blits, outline):
     # The y coordinate of the bottom of the current line.
     bottom_y = 0
 
-
     # The maximum x coordinate of the previous blit on this line.
     max_x = 0
 
@@ -288,7 +288,6 @@ class TextSegment(object):
         on the font group.
         """
 
-
         if not isinstance(self.font, font.FontGroup):
             yield (self, s)
             return
@@ -320,6 +319,7 @@ class TextSegment(object):
         fo = font.get_font(self.font, self.size, self.bold, self.italic, 0, self.antialias, self.vertical, self.hinting, layout.oversample)
         return fo.bounds(glyphs, bounds)
 
+
 class SpaceSegment(object):
     """
     A segment that's used to render horizontal or vertical whitespace.
@@ -403,7 +403,15 @@ class DisplayableSegment(object):
         glyph = glyphs[0]
 
         if di.displayable_blits is not None:
-            di.displayable_blits.append((self.d, glyph.x, glyph.y, glyph.time))
+
+            xo, yo = renpy.display.core.place(
+                glyph.width,
+                glyph.ascent,
+                glyph.width,
+                glyph.line_spacing,
+                self.d.get_placement())
+
+            di.displayable_blits.append((self.d, glyph.x + xo, glyph.y + yo, glyph.time))
 
     def assign_times(self, gt, glyphs):
         if self.cps != 0:
@@ -415,6 +423,7 @@ class DisplayableSegment(object):
     def bounds(self, glyphs, bounds, layout):
         return bounds
 
+
 class FlagSegment(object):
     """
     A do-nothing segment that just exists so we can flag the start and end
@@ -433,6 +442,7 @@ class FlagSegment(object):
     def bounds(self, glyphs, bounds, layout):
         return bounds
 
+
 class Layout(object):
     """
     Represents the layout of text.
@@ -582,7 +592,7 @@ class Layout(object):
             self.paragraph_glyphs.append(list(par_glyphs))
 
             if splits_from:
-                textsupport.copy_splits(splits_from.paragraph_glyphs[p_num], par_glyphs)
+                textsupport.copy_splits(splits_from.paragraph_glyphs[p_num], par_glyphs)  # @UndefinedVariable
 
             else:
 
@@ -678,7 +688,7 @@ class Layout(object):
             target_x = self.scale_int(splits_from.size[0] - splits_from.xborder)
             target_y = self.scale_int(splits_from.size[1] - splits_from.yborder)
 
-            textsupport.tweak_glyph_spacing(all_glyphs, lines, target_x - maxx, target_y - y, maxx, y)
+            textsupport.tweak_glyph_spacing(all_glyphs, lines, target_x - maxx, target_y - y, maxx, y)  # @UndefinedVariable
 
             maxx = target_x
             y = target_y
@@ -743,7 +753,6 @@ class Layout(object):
 
                 ts.draw(glyphs, di, self.add_left, self.add_top, self)
 
-
             renpy.display.draw.mutated_surface(surf)
             tex = renpy.display.draw.load_texture(surf)
 
@@ -774,7 +783,6 @@ class Layout(object):
                 renpy.display.to_log.write("     Available: (%d, %d) Laid-out: (%d, %d)", width, height, sw, sh)
                 renpy.display.to_log.write("     Text: %r", text.text)
 
-
     def scale(self, n):
         if n is None:
             return n
@@ -799,7 +807,6 @@ class Layout(object):
 
         return n * int(self.oversample)
 
-
     def unscale_pair(self, x, y):
         return x / self.oversample, y / self.oversample
 
@@ -841,14 +848,12 @@ class Layout(object):
 
         def fill_empty_line():
             for i in line:
-                if isinstance(i, (TextSegment, SpaceSegment, DisplayableSegment)):
+                if isinstance(i[0], (TextSegment, SpaceSegment, DisplayableSegment)):
                     return
 
             line.extend(tss[-1].subsegment(u" "))
 
-
-
-        for type, text in tokens: #@ReservedAssignment
+        for type, text in tokens:  # @ReservedAssignment
 
             if type == PARAGRAPH:
 
@@ -989,6 +994,17 @@ class Layout(object):
             elif tag == "color":
                 push().color = renpy.easy.color(value)
 
+            elif tag == "alpha":
+                ts = push()
+                if value[0] in "+-":
+                    value = ts.color.alpha + float(value)
+                elif value[0] == "*":
+                    value = ts.color.alpha * float(value[1:])
+                else:
+                    value = float(value)
+
+                ts.color = ts.color.replace_opacity(value)
+
             elif tag == "k":
                 push().kerning = self.scale(float(value))
 
@@ -1053,7 +1069,6 @@ class Layout(object):
 
         return l, rtl
 
-
     def figure_outlines(self, style):
         """
         Return a list containing the outlines, including an outline
@@ -1078,7 +1093,6 @@ class Layout(object):
             for dsx, dsy in dslist:
                 outlines.append((0, style.drop_shadow_color, self.scale_outline(dsx), self.scale_outline(dsy)))
 
-
         for size, color, xo, yo in style_outlines:
             outlines.append((self.scale_outline(size), color, self.scale_outline(xo), self.scale_outline(yo)))
 
@@ -1111,7 +1125,6 @@ class Layout(object):
 
         return outlines, right - left, bottom - top, -left, -top
 
-
     def blits_typewriter(self, st):
         """
         Given a st and an outline, returns a list of blit objects that
@@ -1154,7 +1167,6 @@ class Layout(object):
         min_x = width
         max_x = 0
 
-
         left = False
         right = False
 
@@ -1210,7 +1222,6 @@ virtual_layout_cache_old = { }
 virtual_layout_cache_new = { }
 
 
-
 def layout_cache_clear():
     """
     Clears the old and new layout caches.
@@ -1241,6 +1252,7 @@ def layout_cache_tick():
 VERT_REVERSE = renpy.display.render.Matrix2D(0, -1, 1, 0)
 VERT_FORWARD = renpy.display.render.Matrix2D(0, 1, -1, 0)
 
+
 class Text(renpy.display.core.Displayable):
 
     """
@@ -1270,6 +1282,8 @@ class Text(renpy.display.core.Displayable):
     __version__ = 4
 
     _uses_scope = True
+    _duplicatable = False
+    locked = False
 
     def after_upgrade(self, version):
 
@@ -1343,6 +1357,31 @@ class Text(renpy.display.core.Displayable):
         # The list of displayables we use.
         self.displayables = None
 
+        self._duplicatable = self.slow
+
+    def _duplicate(self, args):
+        if self._duplicatable:
+            rv = self._copy(args)
+            rv.unique()
+
+            return rv
+
+        return self
+
+    def _in_current_store(self):
+        if not self._uses_scope:
+            return self
+
+        rv = self._copy()
+
+        if rv.displayables is not None:
+            rv.displayables = [ i._in_current_store() for i in rv.displayables ]
+
+        rv.slow_done = None
+        rv.locked = True
+
+        return rv
+
     def __unicode__(self):
         s = ""
 
@@ -1362,10 +1401,16 @@ class Text(renpy.display.core.Displayable):
         Called to update the scope, when necessary.
         """
 
+        if self.locked:
+            return False
+
         return self.set_text(self.text_parameter, scope, self.substitute, update)
 
     def set_text(self, text, scope=None, substitute=False, update=True):
 
+        if self.locked:
+            return
+
         old_text = self.text
 
         if not isinstance(text, list):
@@ -1384,7 +1429,10 @@ class Text(renpy.display.core.Displayable):
                     i, did_sub = renpy.substitutions.substitute(i, scope, substitute)
                     uses_scope = uses_scope or did_sub
 
-                i = unicode(i)
+                if isinstance(i, str):
+                    i = unicode(i, "utf-8", "replace")
+                else:
+                    i = unicode(i)
 
             new_text.append(i)
 
@@ -1459,7 +1507,7 @@ class Text(renpy.display.core.Displayable):
         # Tokenize the text.
         tokens = self.tokenize(text)
 
-        if renpy.config.custom_text_tags:
+        if renpy.config.custom_text_tags or (renpy.config.replace_text is not None):
             tokens = self.apply_custom_tags(tokens)
 
         # self.tokens is a list of pairs, where the first component of
@@ -1470,7 +1518,6 @@ class Text(renpy.display.core.Displayable):
         # Text.
         self.tokens, self.displayables = self.get_displayables(tokens)
 
-
     def visit(self):
 
         if self.dirty or self.displayables is None:
@@ -1492,7 +1539,7 @@ class Text(renpy.display.core.Displayable):
         rv = "".join(rv)
         _, _, rv = rv.rpartition("{fast}")
 
-        rv = renpy.translation.notags_filter(rv)
+        rv = renpy.translation.dialogue.notags_filter(rv)
 
         alt = self.style.alt
 
@@ -1517,7 +1564,6 @@ class Text(renpy.display.core.Displayable):
         virtual_layout_cache_old.pop(key, None)
         virtual_layout_cache_new.pop(key, None)
 
-
     def get_layout(self):
         """
         Gets the layout of this text, if one exists.
@@ -1565,7 +1611,6 @@ class Text(renpy.display.core.Displayable):
         if hyperlink_focus:
             return hyperlink_focus(target)
 
-
     def set_style_prefix(self, prefix, root):
         if prefix != self.style.prefix:
             self.kill_layout()
@@ -1616,7 +1661,7 @@ class Text(renpy.display.core.Displayable):
                 return rv
 
         if (self.is_focused() and
-            renpy.display.behavior.map_event(ev, "button_select")):
+                renpy.display.behavior.map_event(ev, "button_select")):
 
             clicked = self.style.hyperlink_functions[1]
 
@@ -1646,6 +1691,21 @@ class Text(renpy.display.core.Displayable):
 
         return layout.unscale_pair(*layout.size)
 
+    def get_time(self):
+        """
+        Returns the amount of time, in seconds, it will take to display this
+        text.
+        """
+
+        layout = self.get_layout()
+
+        # If we haven't been laid out, either the text isn't being shown,
+        # or it's not animated
+        if layout is None:
+            return 0.0
+
+        return layout.max_time
+
     def render(self, width, height, st, at):
 
         if self.style.vertical:
@@ -1767,13 +1827,12 @@ class Text(renpy.display.core.Displayable):
                 else:
                     b_y += layout.add_top
 
-
                 # Blit.
                 rv.absolute_blit(
                     tex.subsurface((b_x, b_y, b_w, b_h)),
                     layout.unscale_pair(b_x + xo + layout.xoffset - o - layout.add_left,
-                     b_y + yo + layout.yoffset - o - layout.add_top)
-                     )
+                                        b_y + yo + layout.yoffset - o - layout.add_top)
+                    )
 
         # Blit displayables.
         if layout.displayable_blits:
@@ -1818,7 +1877,6 @@ class Text(renpy.display.core.Displayable):
 
         return rv
 
-
     def tokenize(self, text):
         """
         Convert the text into a list of tokens.
@@ -1842,7 +1900,6 @@ class Text(renpy.display.core.Displayable):
 
         return tokens
 
-
     def apply_custom_tags(self, tokens):
         """
         Apply new-style custom text tags.
@@ -1955,4 +2012,3 @@ language_tailor = textsupport.language_tailor
 
 # Compatibility, in case one of these was pickled.
 ParameterizedText = extras.ParameterizedText
-
